#-----------------------------------------------------------------------
# Funções para a construção do chatbot Grão Guru.
# Autor: Wagner Hugo Bonat · Insight Technologies

#-----------------------------------------------------------------------
# Pacotes e variáveis de ambiente.

# Pacotes devem ser usados com a declaração do namespace.
##' library(dplyr)
##' library(stringr)
##' library(httr)
##' library(openai)
##' library(lsa)
##' library(lubridate)

# Pacotes não disponíveis irão retornar erro aqui.
find.package("dplyr")
find.package("stringr")
find.package("httr")
find.package("openai")
find.package("lsa")
find.package("lubridate")

# OpenAI API key.
if (Sys.info()["nodename"] == "wz.vps") {
    # readRenviron("/srv/shiny-server/tourchat-dev/.Renviron")
    readRenviron(".Renviron")
}
token <- Sys.getenv("OPENAI_API")
stopifnot(token != "")

#-----------------------------------------------------------------------
# Para limpar HTML.

# Converte anotações HTML `&TEXT;` para texto.
html_decode <- function(txt) {
  XML::htmlParse(txt, asText = TRUE) |>
    XML::getNodeSet("//p") |>
    unlist(recursive = FALSE) |>
    XML::xmlValue()
}
##' txt <- "de execu&ccedil;&atilde;o"
##' html_decode(txt)
##' rm(txt)

# Limpa arquivos HTML.
clean_html <- function(txt, pdf = FALSE) {
  if(pdf) {
    clean_text <- str_replace_all(txt,
                                  pattern = "<.*?>",
                                  replacement = "")
    clean_text <- str_replace_all(clean_text,
                                  pattern = "\\r|\\n|&ndash|&mdash|&#39;",
                                  replacement = " ")
  } else {
    clean_text <- html_decode(txt = txt) |>
      gsub(pattern = "[[:space:]]+", replacement = " ") |>
      trimws()
  }
  return(clean_text)
}

#-----------------------------------------------------------------------
# Para preparar texto para consulta.

# Corta o texto em chunks para submeter para o OpenAI.
organize_document <- function(text,
                              chunk_size,
                              chunk_overlap,
                              title,
                              date,
                              id,
                              pdf = FALSE,
                              token) {
  # Limpa o texto antes de cortar.
  text <- paste(clean_html(text, pdf = pdf), collapse = " ")
  # Testando se o texto tem figuras ou tabelas.
  if (any(grepl("<v:group coordsize=", text) | grepl("v:shape", text))) {
    words <- stringr::str_split(text, "\\s+")[[1]]
    n_characteres <- sapply(words, nchar)
    remove_character <- n_characteres[n_characteres > 50]
    words <- words[!words %in% names(remove_character)]
  } else {
    words <- stringr::str_split(text, "\\s+")[[1]]
  }
  n_words <- length(words)
  if (n_words < chunk_size) {
    n_chunks <- 1
  } else {
    n_chunks <- floor(n_words/chunk_size) + 10
  }
  indexes <- matrix(NA, ncol = 2, nrow = n_chunks)
  indexes[1, ] <- c(1, chunk_size)
  if (n_words > chunk_size) {
    for (i in 2:n_chunks) {
      idx1 <- indexes[c(i - 1), 2] - chunk_overlap
      idx2 <- indexes[c(i - 1), 2] - chunk_overlap + chunk_size
      if (idx2 > n_words) {
        idx2 <- n_words
      }
      indexes[i, ] <- c(idx1, idx2)
      if (idx2 == n_words) {
        break
      }
    }
    indexes <- na.exclude(indexes)
  }
  # Loop para criar os chunks
  n_chunks <- dim(indexes)[1]
  chunks <- data.frame(
    "title" = NA,
    "date" = ymd("2024-01-01"),
    "content" = NA,
    "n_words" = NA,
    "embedding" = NA,
    "n_tokens" = NA,
    "part" = NA,
    "id" = NA
  )

  for (i in 1:n_chunks) {
    # Calcula os índices de início e fim do chunk atual
    # Extrai o chunk atual
    temp_words <- na.exclude(words[indexes[i, 1]:indexes[i, 2]])
    chunks[i, ]$title <- title
    chunks[i, ]$date <- date
    chunks[i, ]$content <- paste(temp_words, collapse = " ")
    chunks[i, ]$n_words <- length(temp_words)
    # Criando o embedding.
    emb <- openai::create_embedding(
      model = "text-embedding-3-small",
      input = chunks[i, ]$content,
      openai_api_key = token
    )
    chunks[i, ]$n_tokens <- emb$usage$total_tokens
    chunks[i, ]$embedding <- emb$data$embedding
    chunks[i, ]$part <- i
    chunks[i, ]$id <- id
  }
  return(chunks)
}

# Cria o vector store baseado na base de conhecimento. O data set de
# conteúdo deve ter três colunas: Content, title e PublishDate.
create_vector_store <- function(context,
                                chunk_size,
                                chunk_overlap,
                                token,
                                pdf = FALSE,
                                verbose = getOption("grao.guru.verbose", FALSE)) {
  n_docs <- dim(context)[1]
  docs <- list()
  for (i in 1:n_docs) {
    docs[[i]] <-
      organize_document(
        text = context$Content[i],
        chunk_size = chunk_size,
        chunk_overlap = chunk_overlap,
        title = context$title[i],
        date = context$PublishDate[i],
        id = context$id[i],
        pdf = pdf,
        token = token
      )
    if (verbose) {
      print(i)
    }
  }
  vector_store <- do.call(rbind, docs)
  return(vector_store)
}

# Chama o chatGPT dado um prompt.
request_ChatGPT_api <- function(prompt, token, temperature = 0.1) {
    chat_GPT_answer <- httr::POST(
        url = "https://api.openai.com/v1/chat/completions",
        httr::add_headers(Authorization = paste("Bearer", token)),
        httr::content_type_json(),
        encode = "json",
        body = list(
            model = "gpt-3.5-turbo", temperature = temperature,
            messages = list(
                list(
                    role = "user",
                    content = prompt
                )
            )
        )
    )
    stringr::str_trim(
        httr::content(chat_GPT_answer)$choices[[1]]$message$content)
}
##' res <- request_ChatGPT_api(
##'     prompt = "Quando o jogo de futebol foi inventado?",
##'     token = Sys.getenv("OPENAI_API"))
##' res |>
##'     strwrap(width = 72) |>
##'     cat(sep = "\n", "\n")
##' rm(res)

#-----------------------------------------------------------------------
# Verifica expressões temporais.

# Aqui a ideia é ir aprendendo como as pessoas perguntam e ir
# melhorando. Essas funções só são aplicadas no prompt inicial.


extract_date <- function(string) {
  padrao <- "\\b\\d{2}/\\d{2}/\\d{4}\\b"
  datas_temp <- stringr::str_extract_all(string, padrao)[[1]]
  return(datas_temp)
}

verify_today <- function(string) {
    # Sinônimos em inglês e português
    sinonimos <- c(
        "today",
        "nowadays",
        "presently",
        "currently",
        "right now",
        "hoje",
        "atual",
        "atualmente",
        "hoje mesmo",
        "neste dia",
        "na atualidade",
        "no dia de hoje",
        "neste momento"
    )
    # Verifica se algum dos sinônimos está presente na string.
    for (sinonimo in sinonimos) {
        if (grepl(sinonimo, string, ignore.case = TRUE)) {
            return(TRUE)
        }
    }
    return(FALSE)
}
##' verify_today("Hoje é um dia de sol.")
##' verify_today("Neste momento a temperatura é de 30 graus.")

verify_yesterday <- function(string) {
    # Sinônimos em inglês e português para "ontem".
    sinonimos <- c(
        "yesterday",
        "the day before",
        "last day",
        "prior day",
        "ontem",
        "ontem mesmo",
        "no dia anterior",
        "no último dia",
        "dia anterior"
    )
    # Verifica se algum dos sinônimos está presente na string
    for (sinonimo in sinonimos) {
        if (grepl(sinonimo, string, ignore.case = TRUE)) {
            return(TRUE)
        }
    }
    return(FALSE)
}
##' verify_yesterday("Ontem foi um dia chuvoso.")

verify_week <- function(string) {
    # Sinônimos em inglês e português para "última semana".
    sinonimos <- c(
        "last week",
        "the past week",
        "previous week",
        "semana passada",
        "últimos sete dias",
        "semana anterior",
        "semana",
        "esta semana",
        "última semana"
    )
    # Verifica se algum dos sinônimos está presente na string.
    for (sinonimo in sinonimos) {
        if (grepl(sinonimo, string, ignore.case = TRUE)) {
            return(TRUE)
        }
    }
    return(FALSE)
}
##' verify_week("Na semana passada choveu muito.")
##' verify_week("Nesta última semana houve aumento na inflação.")

verify_month <- function(string) {
    # Sinônimos em inglês e português para "mês anterior".
    sinonimos <- c(
        "mês anterior",
        "este mês",
        "mês",
        "mês passado",
        "mês passada",
        "último mês",
        "mês que passou",
        "mês findo",
        "mês pretérito",
        "last month",
        "previous month",
        "month prior",
        "month gone by",
        "preceding month"
    )
    # Verifica se algum dos sinônimos está presente na string.
    for (sinonimo in sinonimos) {
        if (grepl(sinonimo, string, ignore.case = TRUE)) {
            return(TRUE)
        }
    }
    return(FALSE)
}
##' verify_month("No mês anterior houve aumento na inflação.")
##' verify_month("No último mês houve aumento na inflação.")

verify_product <- function(string) {
    produtos <- c(
        "produto: morning_wrap_up",
        "produto: mercado_asia",
        "produto: tour_semana",
        "produto: fertilizantes",
        "produto: alertas_mercado"
    )
    opts <- c("morning_wrap_up", "mercado_asia", "tour_semana", "fertilizantes", "alertas_mercado")
    output <- opts[stringr::str_detect(string, produtos)]
    return(output)
}
##' verify_product("Consulte os arquivos sobre 'produto: morning_wrap_up'.")
##' verify_product("Consulte os arquivos sobre 'produto: morning_wrap_up' e 'produto: mercado_asia'.")

# Função auxiliar para verificar se um período de datas esta contindo em
# outro.
check_period_within <- function(inner_start,
                                inner_end,
                                outer_start,
                                outer_end) {
    # Convertendo as datas para o formato Date.
    inner_start <- as.Date(inner_start, format = "%d/%m/%Y")
    inner_end <- as.Date(inner_end, format = "%d/%m/%Y")
    outer_start <- as.Date(outer_start, format = "%d/%m/%Y")
    outer_end <- as.Date(outer_end, format = "%d/%m/%Y")

    # Verificando se o período 1 está dentro do período 2.
    if (outer_start <= inner_start & inner_end <= outer_end) {
        return(TRUE)
    } else {
        return(FALSE)
    }
}
##' check_period_within("01/01/2022", "31/01/2022",
##'                     "01/01/2022", "31/01/2022")
##' check_period_within("10/01/2022", "20/01/2022",
##'                     "01/01/2022", "31/01/2022")
##' check_period_within("10/01/2022", "20/01/2022",
##'                     "15/01/2022", "31/01/2022")

# Função para criar os filtros de metadados.
get_metadata <- function(prompt_user, date_min, date_max, token) {

   #prompt_extrai_data <- paste("Extraia o período de data relevante para responder ao
   #                             prompt do usuário fornecido entre colchetes.
   #                            Organize a data no formato DD/MM/AAAA.")
    prompt_extrai_data <- paste(
        "Atue como um extrator de dados. Baseado no texto entre colchetes",
        "extraia e organize no formato DD/MM/AAAA a data inicial e a",
        "data final relevante para responder a pergunta do usuário.",
        "Apenas retorne a data se as encontrar no texto entre colchetes.",
        "Retorne apenas as datas no formato de código R para criar um",
        "vetor onde a primeira posição é a data inicial e a segunda",
        "posição é a data final. Use este exemplo: DD/MM/AAAA DD/MM/AAAA.")

    datas_str <- c()

    # Testando se tem algo sobre HOJE.
    if (verify_today(prompt_user)) {
        captura_data <-
            format(as.Date(Sys.Date(), format = "%d-%m-%Y"),
                   "%d/%m/%Y")
        datas_str <- c(datas_str, captura_data)
    }
    # Testando se tem algo sobre ONTEM.
    if (verify_yesterday(prompt_user)) {
        captura_data <-
            format(as.Date(Sys.Date() - 1, format = "%d-%m-%Y"),
                   "%d/%m/%Y")
        datas_str <- c(datas_str, captura_data)
    }
    # Testando se tem algo com SEMANA.
    if (verify_week(prompt_user)) {
        data_ultima <-
            format(as.Date(Sys.Date() - 7, format = "%d-%m-%Y"),
                   "%d/%m/%Y")
        data_atual <-
            format(as.Date(Sys.Date(), format = "%d-%m-%Y"),
                   "%d/%m/%Y")
        datas_str <- c(datas_str, data_ultima, data_atual)
    }
    # Testando se tem algo com MÊS.
    if (verify_month(prompt_user)) {
        data_ultima <-
            format(as.Date(Sys.Date() - 30, format = "%d-%m-%Y"),
                   "%d/%m/%Y")
        data_atual <-
            format(as.Date(Sys.Date(), format = "%d-%m-%Y"),
                   "%d/%m/%Y")
        datas_str <- c(datas_str, data_ultima, data_atual)
    }

    ## Verificando se tem datas especificas
    datas_temporaria <- extract_date(prompt_user)
    if(length(datas_temporaria) != 0) {
      datas_str <- c(datas_str, datas_temporaria)
    }
    if (!is.null(datas_str)) {
        data_inicial <- min(lubridate::dmy(datas_str))
        data_final <- max(lubridate::dmy(datas_str))
    }
    ## Verificando se tem menção a uma data específica
    if (is.null(datas_str)) {
        prompt_completo <-
            paste(prompt_extrai_data,
                  paste("[", prompt_user, "]"))
        busca_data <-
            request_ChatGPT_api(prompt = prompt_completo,
                                token = token)
        padrao <- "\\b\\d{2}/\\d{2}/\\d{4}\\b"
        datas_temp <- stringr::str_extract_all(busca_data, padrao)[[1]]
        if (length(datas_temp) != 0) {
            if (check_period_within(min(datas_temp),
                                    max(datas_temp),
                                    date_min - 20,
                                    date_max + 30)) {
                data_inicial <- min(lubridate::dmy(datas_temp))
                data_final <- max(lubridate::dmy(datas_temp))
            } else {
                data_inicial <- date_min
                data_final <- date_max
            }
        } else {
            data_inicial <- date_min
            data_final <- date_max
        }
    }
    # Verificando possíveis problemas.
    if (is.infinite(data_inicial)) {
        data_inicial <- date_min
    }
    if (is.infinite(data_final)) {
        data_final <- date_max
    }
    # Capturando se tem algo sobre um produto especifico.
    produto <- verify_product(prompt_user)
    if (length(produto) == 0) {
        produto <- c("morning_wrap_up", "mercado_asia", "fertilizantes", "alertas_mercado")
    }
    output <- list()
    output$date <- c(
        as.Date(data_inicial, format = "%d/%m/%Y"),
        as.Date(data_final, format = "%d/%m/%Y")
    )
    output$product <- produto
    return(output)
}
##' res <- get_metadata(
##'     prompt_user = "Consulte os arquivos sobre 'produto: morning_wrap_up'.",
##'     date_min = as.Date("01/01/2022", format = "%d/%m/%Y"),
##'     date_max = as.Date("31/01/2022", format = "%d/%m/%Y"),
##'     token = Sys.getenv("OPENAI_API"))
##' # dput(res)
##' # list(date = structure(c(18993, 19023), class = "Date"),
##' #      product = "morning_wrap_up")

#-----------------------------------------------------------------------
# Função que cria a conversação (chat).

##' prompt_chat_history <- "Given a chat history between brackets and the latest user question which might reference context in the chat history, formulate a standalone question which can be understood without the chat history. Do NOT answer the question, just reformulate it if needed and otherwise return it as is."
##' prompt_chat_history |>
##'     str_wrap(width = 50) |>
##'     cat("\n")

# prompt_chat_history <- "Dado o contexto de uma conversa reformule a questão usando o contexto de modo que a pergunta possa ser usada de forma independente. Não responsa a questão, apenas reformule se necessário."
##' prompt_chat_history |>
##'     str_wrap(width = 50) |>
##'     cat("\n")

grao_guru <- function(chat_history,
                      vector_store,
                      token,
                      verbose = getOption("grao.guru.verbose", FALSE)) {
    prompt_assistant <- "You are an assistant for question-answering tasks. Use the following pieces of retrieved context to answer the question. If you don't know the answer, just say that you don't know. Strucuture your answer in topic format. Use numbers to base your answer."
    #prompt_assistant <- "Você é um assistente para tarefas de resposta a perguntas. Use as seguintes partes de contexto recuperado para responder à pergunta. Se você não sabe a resposta, basta dizer que não sabe. Estruture sua resposta em formato de tópico. Use números para basear sua resposta."
    #prompt_chat_history <- "Dado o contexto de uma conversa reformule a questão usando o contexto de modo que a pergunta possa ser usada de forma independente. Não responsa a questão, apenas reformule se necessário."
    prompt_chat_history <- "Given a chat history between brackets and the latest user question between brackets which might reference to the context in the chat history, formulate a standalone question which can be understood without the chat history. Do NOT answer the question, just reformulate it if needed and otherwise return it as is."
    ##' prompt_assistant |>
    ##'     str_wrap(width = 50) |>
    ##'     cat("\n")

    human_messages <- chat_history |>
        dplyr::filter(role == "user")
    last_message <- dim(human_messages)[1]
    # if(last_message > 1) {
    #   prompt_user <- paste(human_messages$prompt, collapse = " ")
    #   prompt <- paste(
    #     "instructions:",
    #     prompt_chat_history,
    #     "chat history:",
    #     "[",
    #     paste(human_messages$prompt[c(last_message - 1)],
    #           collapse = " "),
    #     "]",
    #     "lastest user question:",
    #     "[",
    #     human_messages$prompt[last_message],
    #     "]",
    #     collapse = " "
    #   )
    #   prompt_user <- request_ChatGPT_api(prompt, token = token)
    # }
    prompt_user <- human_messages$prompt[last_message]
    filtros <- get_metadata(
        prompt_user = prompt_user,
        date_min = min(vector_store$date),
        date_max = max(vector_store$date),
        token = token
    )
    ##' str(filtros)

    # Selecionando os documentos que interessam de acordo com os
    # filtros.
    conhecimento <- vector_store |>
        dplyr::filter(date >= filtros$date[1] &
                          date <= filtros$date[2] &
                              Produto %in% filtros$product)
    conhecimento_fertilizantes <- vector_store |>
      dplyr::filter(date >= filtros$date[2] - 6 &
                      Produto == "fertilizantes")
    conhecimento <- rbind(conhecimento, conhecimento_fertilizantes)


    ##' str(conhecimento, max.level = 1)

    # Verificando quantos token tem na base de conhecimento.
    n_tokens <- sum(conhecimento$n_tokens)

    if (n_tokens > 4000) {
        prompt_user_embedding <- openai::create_embedding(
            model = "text-embedding-3-small",
            input = prompt_user,
            openai_api_key = token
        )

        similaridade <- c()
        for (i in 1:dim(conhecimento)[1]) {
            similaridade[i] <- lsa::cosine(
                prompt_user_embedding$data$embedding[[1]],
                y = conhecimento$embedding[[i]]
            )
        }
        # Trazendo apenas os documentos mais relevantes de acordo com o
        # prompt.
        conhecimento$similaridade <- similaridade
        conhecimento <- conhecimento |>
            dplyr::arrange(dplyr::desc(similaridade))

        conhecimento$token_acumulado <- cumsum(conhecimento$n_tokens)
        conhecimento <- conhecimento |>
            dplyr::filter(token_acumulado < 3900)
        contexto <- paste(conhecimento$content, collapse = " ")
    } else {
        contexto <- paste(conhecimento$content, collapse = " ")
    }

    # str(conhecimento, max.level = 1)
    # str(contexto, max.level = 1)
    print(prompt_user)
    prompt <- paste(
        paste("system:", prompt_assistant),
        paste("context:", contexto),
        paste("user question:", prompt_user)
    )
    saida <- request_ChatGPT_api(prompt, token = token)

    if (verbose) {
        saida |>
            strwrap(width = 60) |>
            cat(sep = "\n")
    }

    chat_history <- rbind(chat_history,
                          data.frame("role" = "ai",
                                     "prompt" = saida))
    data_busca <- filtros$date
    produto <- filtros$product
    output <- list(
        "chat_history" = chat_history,
        "date_search" = data_busca,
        "product" = produto,
        "documents" = conhecimento
    )
    return(output)
}

##' # Carregando a base de dados.
##' load("vector_store.RData")
##' stopifnot(exists("vector_store"))
##' str(vector_store, max.level = 1)
##'
##' # Faz uma chamada para o Grão Guru.
##' prompt <- "O que é farmer selling?"
##' prompt <- "Como foi o mercado de soja semana passada?"
##' prompt <- "Me fale sobre o número de navios de soja que a China comprou nos últimos 15 dias."
##' chat_history <-
##'     create_chat_history(prompt = prompt,
##'                         role = "user")
##' answer <-
##'     grao_guru(chat_history = chat_history,
##'               vector_store = vector_store,
##'               token = Sys.getenv("OPENAI_API"))
##' str(answer, max.level = 1)
##' str(answer$documents, max.level = 1)
##'
##' tail(answer$chat_history$prompt, n = 1) |>
##'     strwrap(width = 60) |>
##'     cat(sep = "\n", "")
##'
##' for (i in 1:nrow(answer$documents)) {
##'     cat("------------------------------------------------------------\n")
##'     answer$documents$title[i] |>
##'         strwrap(width = 60) |>
##'         cat(sep = "\n")
##'     cat("Similaridade:",
##'         answer$documents$similaridade[1],
##'         "\n\n")
##'     answer$documents$similaridade[i]
##'     answer$documents$content[i] |>
##'         strwrap(width = 60) |>
##'         cat(sep = "\n")
##'     cat("\n")
##' }
##'
##' # Fonte da notícia.
##' url <- "https://go.agrinvest.agr.br/noticias/details/"
##' browseURL(paste0(url, answer$documents$id[1]))
##'
##' answer$chat_history$prompt |>
##'     tail(n = 1) |>
##'     format_message(emoji = "🤖") |>
##'     cat(sep = "\n")


#-----------------------------------------------------------------------
# Para facilitar a manipulação do chat_history.

# Função para facilitar pegar a última resposta no chat_history.
get_last_answer <- function(chat_history) {
    IA <- chat_history |>
        dplyr::filter(role == "ai")
    n_last <- dim(IA)[1]
    return(IA$prompt[n_last])
}

update_chat_history <- function(prompt, role, chat_history) {
    update <- data.frame("role" = role, "prompt" = prompt)
    chat_history <- rbind(chat_history, update)
    return(chat_history)
}

create_chat_history <- function(prompt, role) {
    chat_history <- data.frame("role" = role, "prompt" = prompt)
    return(chat_history)
}

#-----------------------------------------------------------------------
# Para testar o chat em ambiente de console.

format_message <- function(..., emoji = " 🤖") {
    message <- paste(emoji, ...) |>
        # strwrap(width = 60, exdent = 4) |>
        strwrap(width = 60, exdent = 3, prefix = "| ")
    # message <- paste(emoji, "\n", message)
    return(message)
}

tourchat_console <- function(vector_store) {
    emoji <- c("🤖", "🙂")
    "Bem-vindo(a) ao Grão Guru!" |>
        format_message(emoji = emoji[1]) |>
        cat(sep = "\n", "\n")
    "Sou o assistente virtual da Agrinvest e vou ajudar você a ficar por dentro das notícias." |>
        format_message(emoji = emoji[1]) |>
        cat(sep = "\n", "\n")
    "Vamos começar?" |>
        format_message(emoji = emoji[1]) |>
        cat(sep = "\n")
    while (TRUE) {
        prompt <- "Escreva uma pergunta e pressione ENTER:\n" |>
            format_message(emoji = emoji[1]) |>
            paste("\n\n| 🙂 ") |>
            readline()
        chat_history <-
            create_chat_history(prompt = prompt,
                                role = "user")
        saida_gg <-
            grao_guru(chat_history = chat_history,
                      vector_store = vector_store,
                      token = Sys.getenv("OPENAI_API"))
        cat("\n")
        saida_gg$chat_history$prompt |>
            tail(n = 1) |>
            format_message(emoji = emoji[1]) |>
            cat(sep = "\n")
        cat("\n")
    }
}

##' # Carregando a base de dados.
##' load("vector_store.RData")
##' stopifnot(exists("vector_store"))
##' str(vector_store, max.level = 1)
##'
##' # Exemplos de prompt.
##' # "Quantos barcos de soja a China comprou na semana passada?"
##' # "Por favor, elabore mais sobre a compra de navios de soja pela China."
##' # "O que é farmer selling?"
##'
##' # Inicia o chat para conversar no console.
##' tourchat_console(vector_store)
##'
##' # Faz uma chamada para o Grão Guru.
##' prompt <- "O que é farmer selling?"
##' chat_history <-
##'     create_chat_history(prompt = prompt,
##'                         role = "user")
##' answer <-
##'     grao_guru(chat_history = chat_history,
##'               vector_store = vector_store,
##'               token = Sys.getenv("OPENAI_API"))
##' str(answer, max.level = 1)
##' str(answer$documents, max.level = 1)
##'
##' # Fonte da notícia.
##' url <- "https://go.agrinvest.agr.br/noticias/details/"
##' browseURL(paste0(url, answer$documents$id[1]))
##'
##' answer$chat_history$prompt |>
##'     tail(n = 1) |>
##'     format_message(emoji = "🤖") |>
##'     cat(sep = "\n")

#-----------------------------------------------------------------------
