#-----------------------------------------------------------------------
# Carregando as funções do arquivo `functions.R` com alinas `wb`.

box::use(wb = ./functions)
# box::reload(wb)

# FALSE para não imprimir mensagens de log.
options(grao.guru.verbose = FALSE)

#----------------------------------------------------------------------
# Verifica se existe uma variável de ambiente chamada `OPENAI_API`.

if (Sys.getenv("OPENAI_API") == "") {
    stop("A variável de ambiente `OPENAI_API` não foi encontrada.")
}

#-----------------------------------------------------------------------
# Verifica se as variáveis de ambiente estão definidas.

# Variáveis de ambiente estão definidas em arquivos não versionados.
# Verifique se estão no `.Renviron` ou `.Rprofile`.
for (va in c("DBNAME", "HOST", "PORT", "USER", "PASSWORD")) {
    if (Sys.getenv(va) == "") {
        stop(sprintf("Variável de ambiente `%s` não definida. Confira o `.Renviron` ou `.Rprofile`", va))
    }
}

#-----------------------------------------------------------------------
# https://github.com/dreamRs/shinyWidgets/blob/26838f9e9ccdc90a47178b45318d110f5812d6e1/R/useBs4Dash.R

#' Use 'bs4Dash' in 'shiny'
#'
#' Allow to use functions from 'bs4Dash' into a classic 'shiny' app,
#' specifically `bs4ValueBox`, `bs4InfoBox` and `bs4Card`.
#'
#' @export
#'
#' @param ... Not used.
#'
#' @importFrom htmltools findDependencies attachDependencies
#'
#' @example examples/useBs4Dash.R
useBs4Dash <- function(...) {
    if (!requireNamespace(package = "bs4Dash"))
        message("Package 'bs4Dash' is required to run this function")
    deps <- htmltools::findDependencies(bs4Dash::bs4DashPage(
        header = bs4Dash::bs4DashNavbar(),
        sidebar = bs4Dash::bs4DashSidebar(),
        body = bs4Dash::bs4DashBody()
    ))
    htmltools::attachDependencies(tags$div(), value = deps)
}

#-----------------------------------------------------------------------
# Carregando a base de dados.

file.exists("vector_store.RData") |>
    stopifnot()

load("vector_store.RData")

stopifnot(exists("vector_store"))
# str(vector_store, max.level = 1)
# dput(names(vector_store))
# dput(vector_store[0, ])

# Isso é como o `vector_store` deve ser.
vector_store_template <-
    structure(
        list(title = character(0),
             date = structure(numeric(0), class = "Date"),
             content = character(0),
             n_words = integer(0),
             embedding = list(),
             n_tokens = integer(0),
             part = integer(0),
             id = integer(0),
             Produto = character(0)),
        row.names = character(0),
        class = "data.frame"
    )

if (!all.equal(vector_store[0, ], vector_store_template)) {
    stop("O objeto `vector_store` não está no formato correto.")
}

#-----------------------------------------------------------------------

# Primeira mensagem do chat dando as boas-vindas.
welcome_user <-
    paste(
        "Olá! Seja bem-vindo(a). Eu sou a IA da Agrinvest.",
        "Estou aqui para ajudar você a ficar atualizado(a)",
        "com as últimas notícias do mercado de commodities.",
        "Digite a sua pergunta para começarmos."
    )
welcome_stranger <-
    paste(
        "Olá! Seja bem-vindo(a). Eu sou a IA da Agrinvest.",
        "Estou aqui para ajudar você a ficar atualizado(a)",
        "com as últimas notícias do mercado de commodities.",
        "Conteúdo, verifiquei que você é um(a) usuário(a) <strong>ainda não autenticado</strong>.",
        "Por razões de segurança, sua interação será limitada.",
        "Entre em contato com os administradores para obter acesso."
    )
##' chat_history <-
##'     wb$create_chat_history(prompt = welcome_user,
##'                            role = "ai")
##' wb$update_chat_history(
##'     chat_history = chat_history,
##'     prompt = "Qual a previsão para o preço do milho em 2024?",
##'     role = "user")

initial_chat_history <-
    wb$create_chat_history(prompt = welcome_user,
                           role = "ai")

#-----------------------------------------------------------------------

# Função que formata a data e hora para exibir ao lado das mensagens.
format_time <- function(x = Sys.time()) {
    strftime(x, format = "%d/%b/%Y %H:%M") |>
        sub(pattern = "/([[:lower:]])",
            replacement = "/\\U\\1",
            perl = TRUE,
            x = _)
}
##' format_time()

# Tabela com os metadados dos usuários. As imagens devem estar no
# diretório `./www`.
chat_metadata <- data.frame(
    role = c("user", "ai"),
    name = c("Agrinvest", "Assistant"),
    image = c("./user.svg",
              "./assistant.svg")
)
##' chat_metadata

# Função que cria uma mensagem de chat. Ela atua conforme o papel do
# usuário.
chat_message <- function(...,
                         date = NULL,
                         role = c("user", "ai"),
                         chat_metadata) {
    role <- match.arg(role)
    author <- chat_metadata$name[chat_metadata$role == role]
    image <- chat_metadata$image[chat_metadata$role == role]
    list(
        author = author,
        date = date,
        image = image,
        type = ifelse(role == "user", "sent", "received"),
        ...
    )
}
##' chat_message("Olá, tudo bem?",
##'              role = "ai",
##'              chat_metadata = chat_metadata)
##'
##' chat_message("Eu quero novidades sobre soja.",
##'              role = "h",
##'              chat_metadata = chat_metadata)

adds_links <- function(resp) {
    if (length(resp$documents$id) == 0) {
        return(resp)
    }
    fmt <- "<a href='https://go.agrinvest.agr.br/noticias/details/%s'>%s</a>"
    links <- resp$documents |>
        dplyr::select(id, title) |>
        dplyr::distinct(id, .keep_all = TRUE) |>
        dplyr::mutate(title = sprintf(fmt, id, title)) |>
        dplyr::pull(title) |>
        paste(collapse = "<br>")
    links[1] <- paste("<br><hr><span class='source'>Para saber mais leia:</span><br>",
                      links[1], sep = "")
    links <- links |>
        paste(collapse = "<br>")
    i <- tail(which(resp$chat_history$role == "ai"), n = 1)
    resp$chat_history$prompt[i] <-
        paste(resp$chat_history$prompt[i], links, sep = "<br>")
    return(resp)
}

create_meassages_list <- function(chat_history, chat_metadata) {
    msgs <-
        tapply(chat_history,
               seq_len(nrow(chat_history)),
               function(msg) {
                   chat_message(
                       msg$prompt,
                       date = if (is.null(msg$date)) NULL else msg$date,
                       role = msg$role,
                       chat_metadata = chat_metadata
                   )
               })
    return(msgs)
}
##' chat_history <-
##'     wb$create_chat_history(prompt = welcome_user,
##'                            role = "ai") |>
##'     wb$update_chat_history(
##'         prompt = "Qual a previsão para o preço do milho em 2024?",
##'         role = "user")
##' chat_history
##'
##' create_meassages_list(
##'     chat_history = chat_history,
##'     chat_metadata = chat_metadata
##' )
##'
##' resp <- wb$grao_guru(chat_history = chat_history,
##'                      vector_store = vector_store,
##'                      token = Sys.getenv("OPENAI_API"))
##' str(resp, max.level = 1)
##' resp$chat_history |>
##'     str(max.level = 1)
##'
##' resp <- adds_links(resp)
##' resp$chat_history$prompt |>
##'     tail(n = 1) |>
##'     strwrap(width = 80) |>
##'     cat(sep = "\n")
##'
##' chat_history <-
##'     resp$chat_history |>
##'     wb$update_chat_history(
##'         prompt = "Qual a previsão para o preço da soja em 2024?",
##'         role = "user")
##'
##' resp <- wb$grao_guru(chat_history = chat_history,
##'                      vector_store = vector_store,
##'                      token = Sys.getenv("OPENAI_API"))
##' str(resp, max.level = 1)
##' resp$chat_history |>
##'     str(max.level = 1)
##'
##' chat_history <-
##'     resp$chat_history
##'
##' create_meassages_list(chat_history = chat_history,
##'                       chat_metadata = chat_metadata) |>
##'     bs4Dash::userMessages(width = 6,
##'                           status = "primary",
##'                           id = "MESSAGE_AREA")

# browseURL("https://github.com/mattboldt/typed.js")
# Baixa o arquivo `typed.min.js` se ele não existir.
js_lib <- "www/typed.min.js"
if (!file.exists(js_lib)) {
    dir.create("www", showWarnings = FALSE)
    "https://cdn.jsdelivr.net/npm/typed.js@2.0.11/lib/typed.min.js" |>
        download.file(destfile = js_lib)
}

repair_string <- function(x) {
    # Troca as quebras de linha por <br> e as aspas duplas por \\". Isso
    # é necessário para usar a `typed.js`.
    x |>
        gsub(pattern = "\n", replacement = "<br>") |>
        gsub(pattern = "\"", replacement = '\\\\"')
}

#-----------------------------------------------------------------------

# Função que aplica uma consulta e retorna o resultado.
db_recordset <- function(query) {
    con <- odbc::dbConnect(drv =  RMySQL::MySQL(),
                           user = Sys.getenv("USER"),
                           password = Sys.getenv("PASSWORD"),
                           dbname = Sys.getenv("DBNAME"),
                           port = as.integer(Sys.getenv("PORT")),
                           host = Sys.getenv("HOST"))
    # on.exit(odbc::dbDisconnect(con), add = TRUE)
    rs <- DBI::dbFetch(DBI::dbSendQuery(con, query), n = -1)
    odbc::dbDisconnect(con)
    return(rs)
}
##' db_recordset("SELECT * FROM gg_term")
##' db_recordset("SELECT * FROM gg_term WHERE fk_user = 1")
##' db_recordset("SELECT * FROM gg_term WHERE fk_user = 5000")
##' db_recordset("SELECT * FROM gg_session")
##' db_recordset("SELECT * FROM gg_hist_chat")

# Função que executa a comandos no banco de dados mas não tem retorno.
db_comand <- function(db_comand) {
    con <- odbc::dbConnect(drv =  RMySQL::MySQL(),
                           user = Sys.getenv("USER"),
                           password = Sys.getenv("PASSWORD"),
                           dbname = Sys.getenv("DBNAME"),
                           port = as.integer(Sys.getenv("PORT")),
                           host = Sys.getenv("HOST"))
    # on.exit(odbc::dbDisconnect(con), add = TRUE)
    odbc::dbSendQuery(con, db_comand)
    invisible(odbc::dbDisconnect(con))
}
##' # Inserir um registro.
##' db_comand(sprintf("INSERT INTO gg_term (fk_user) VALUES (%s)",
##'                   1234))
##'
##' # Conferir.
##' db_recordset("SELECT * FROM gg_term WHERE fk_user = 1234")
##'
##' # Deletar um registro.
##' db_comand(sprintf("DELETE FROM gg_term WHERE fk_user = %s",
##'                   1234))
##'
##' # Conferir.
##' db_recordset("SELECT * FROM gg_term WHERE fk_user = 1234")

# Função que escreve novos registros em uma tabela do banco de dados.
db_save_dt <- function(db_table, data_table) {
    con <- odbc::dbConnect(drv =  RMySQL::MySQL(),
                           user = Sys.getenv("USER"),
                           password = Sys.getenv("PASSWORD"),
                           dbname = Sys.getenv("DBNAME"),
                           port = as.integer(Sys.getenv("PORT")),
                           host = Sys.getenv("HOST"))
    # on.exit(odbc::dbDisconnect(con), add = TRUE)
    odbc::dbWriteTable(conn = con ,
                       db_table,
                       data_table,
                       row.names = FALSE,
                       append = TRUE)
    odbc::dbDisconnect(con)
}
##' db_save_dt(db_table = "gg_term",
##'            data_table = data.frame(fk_user = c(666, 777),
##'                                    accept = c(0, 0)))
##' db_comand("DELETE FROM gg_term WHERE accept = 0")
##'
##' hist_chat <- c("Olá, sou o Grão Guru.",
##'                "Grão Guru, isso é um teste.",
##'                "Tudo bem. Fique à vontade.")
##' agente <- c("ai", "user", "ai")
##' db_save_dt(db_table = "gg_hist_chat",
##'            data_table = data.frame(fk_session = -999,
##'                                    message = hist_chat,
##'                                    agente = agente,
##'                                    stringsAsFactors = FALSE))
##'
##' db_comand("DELETE FROM gg_hist_chat WHERE fk_session = -999")

# chat_history <-
#     wb$create_chat_history(prompt = welcome_user,
#                            role = "ai") |>
#     wb$update_chat_history(
#         prompt = "Qual a previsão para o preço da soja em 2024?",
#         role = "user")
# str(chat_history)
#
# resp <- wb$grao_guru(chat_history = chat_history,
#                      vector_store = vector_store,
#                      token = Sys.getenv("OPENAI_API"))
# str(resp, max.level = 1)
# resp$chat_history |>
#     str(max.level = 1)
#
# tail(resp$chat_history, n = 1) |>
#     tail(n = 1) |>
#     strwrap(width = 80) |>
#     cat(sep = "\n")
#
# db_save_dt(db_table =  "gg_hist_chat",
#            data_table = data.frame(fk_session = -999,
#                                    message = resp$chat_history$prompt,
#                                    agente = resp$chat_history$role))
#
# db_comand("DELETE FROM gg_hist_chat WHERE fk_session = -999")

# Função que valida o token passado e salva no banco de dados
# informações da sessão.
check_agritoken <- function(xtoken) {
    # Este é um token para teste.
    # xtoken <- "teste"

    for (va in c("CLIENTID", "SECRETID", "COOKIE")) {
        if (Sys.getenv(va) == "") {
            stop(sprintf("Variável de ambiente `%s` não definida. Confira o `.Renviron` ou `.Rprofile`", va))
        }
    }

    headers <- c(
        "ClientId" = Sys.getenv("CLIENTID"),
        "SecretId" = Sys.getenv("SECRETID"),
        "Content-Type" = "application/json",
        "Cookie" = Sys.getenv("COOKIE")
    )
    body <- sprintf('{ "token": "%s", "source": 1 }', xtoken)

    # Faz a requisição para o servidor verificar a validade do token.
    res <- httr::VERB(
        verb = "POST",
        url = "https://homolog-agrinvest.azurewebsites.net/api/v2/shiny/validate-token",
        body = body,
        httr::add_headers(headers)
    )

    if (res$status_code == 200L) {

        # json_obj <- (httr::content(res, "text"))
        # parsed_obj <- jsonlite::fromJSON(json_obj)
        parsed_obj <- httr::content(res, "parsed")

        # Salva o ID no banco de dados.
        query_insert <-
            stringr::str_interp(
                "INSERT INTO gg_session (fk_client) VALUES ('${fk_client}')",
                list(fk_client = parsed_obj$userId)
            )
        db_comand(query_insert)

        # Retorna o ID da sessão.
        query_get_id <-
            stringr::str_interp(
                "SELECT MAX(id) as id from gg_session where fk_client='${fk_client}'",
                list(fk_client = parsed_obj$userId)
            )
        fk_session <-  db_recordset(query_get_id)

        token_check <-  list(
            id_session = fk_session[[1]],
            status = res$status_code,
            id_user = parsed_obj$userId
        )

    } else {
        token_check <-  list(
            id_session = NULL,
            status = res$status_code
        )
    }

    return(token_check)
}
##' # Reseta o token no banco de dados para teste.
##' sqlcon <- odbc::dbConnect(drv      = odbc::odbc(),
##'                           Driver   = "ODBC Driver 17 for SQL Server",
##'                           Server   = Sys.getenv("TKSERVER"),
##'                           Database = Sys.getenv("TKDATABASE"),
##'                           UID      = Sys.getenv("TKUID"),
##'                           PWD      = Sys.getenv("TKPWD"),
##'                           Port     = as.integer(Sys.getenv("TKPORT")))
##' DBI::dbSendQuery(sqlcon,
##'                  "UPDATE ShinyToken SET ScheduleUsed = 0")
##' DBI::dbDisconnect(sqlcon)
##'
##' check_agritoken("teste")
##' check_agritoken("johndoe")

# Função que retorna TRUE se o usuário aceitou os termos e FALSE caso
# contrário.
terms_accepted <- function(id_user) {
    query <-
        "SELECT CASE
            WHEN EXISTS (
                SELECT 1
                FROM agrinvest_quotes.gg_term
                WHERE fk_user = %s AND accept = '1'
            )
            THEN 1
            ELSE 0
        END AS user_accepted;" |>
        sprintf(id_user)
    return(as.logical(db_recordset(query)$user_accepted))
}
##' terms_accepted(1)
##' terms_accepted(123)
##' terms_accepted(5000)
##' terms_accepted(5002)
